/*============================================================================
  File:     The Tipping Point.sql

  Summary:  Various examples and ways to see the estimated and actual tipping
			point of when a nonclustered index is not used vs. doing a table
			scan.
  
  Date:     November 2009

  SQL Server Version: SQL Server 2005/2008
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SYSolutions, Inc.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE AdventureWorksDW2008
go

sp_helpindex2 FactInternetSales
go

SELECT S.CustomerKey, COUNT(*) AS NumTotalSales
FROM FactInternetSales AS S
GROUP BY S.CustomerKey
ORDER BY NumTotalSales DESC
COMPUTE SUM(COUNT(*))
go

SELECT * FROM sys.dm_db_index_physical_stats
	(db_id(), object_id('FactInternetSales'), NULL, NULL, 'limited')

-- Clustered Index has 1025 PAGES
-- 1/4 = ~256
-- 1/3 = ~338

-- Turn on showplan (Ctrl+K)
SET STATISTICS IO ON
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11048
GROUP BY S.CustomerKey, C.LastName
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11049
GROUP BY S.CustomerKey, C.LastName
go

-- Actual tipping point between 301 and 307...

-- If this query were critical... 
-- I'd first look at the missing index suggestion 
--   (the "green" suggestion in showplan - none for this execution as it's
--    too small/not enough of an improvement BUT for the larger version of this
--    table we get:)

--CREATE NONCLUSTERED INDEX [<Name of Missing Index, sysname,>]
--ON [dbo].[FactInternetSales] ([CustomerKey])
--INCLUDE ([SalesAmount])

-- But... before you create it make sure to FIRST check existing indexes:
-- You can grab sp_helpindex2 here: http://www.sqlskills.com/BLOGS/KIMBERLY/post/Updates-(fixes)-to-sp_helpindex2.aspx

sp_helpindex2 FactInternetSales
go

--1) Create the new index

CREATE NONCLUSTERED INDEX IX_FactInternetSales_CustomerKey_INCSalesAmount
ON [dbo].[FactInternetSales] ([CustomerKey])
INCLUDE ([SalesAmount])
go

--2) Drop the old???
--   we should always check our indexes first to see if any can be dropped/consolidated 
--   based on the new index being created!

DROP INDEX [FactInternetSales].IX_FactInternetSales_CustomerKey
go

-- May need to drop the old before creating the new 
-- (if space is an issue)
-- but this may negatively affect existing users?!

-- Now, no matter what the range and/or number of rows, this index can always be used:
SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11048
GROUP BY S.CustomerKey, C.LastName
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11049
GROUP BY S.CustomerKey, C.LastName
go

-- What about an even bigger range:
SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey > 11049
GROUP BY S.CustomerKey, C.LastName
go

-- Finding the *estimated* tipping point for every table:
USE AdventureWorksDW2008
go

SELECT OBJECT_NAME(object_id) AS [TableName]
	, record_count AS [RowCount]
	, page_count
    , page_count * .25 AS LowRangePages
    , page_count * .33 AS HighRangePages
    , ((page_count * .25)/record_count * 100) AS [LowRangeRows (Percent)]
    , ((page_count * .33)/record_count * 100) AS [HighRangeRows (Percent)]
    , * 
FROM sys.dm_db_index_physical_stats
(db_id(), null, null, null, 'sampled')
WHERE index_id IN (0,1) 
AND page_count > 100
ORDER BY [RowCount] DESC
go

-- What is your database's average?
SELECT DB_NAME(), AVG(tp.[LOW]) AS [Avg Low Percent]
	, AVG(tp.[High]) AS [Avg High Percent]
FROM (SELECT ((page_count * .25)/record_count *100) AS [Low]
    , ((page_count * .33)/record_count * 100) AS [High]
	FROM sys.dm_db_index_physical_stats
		(db_id(), null, null, null, 'sampled')
	WHERE index_id IN (0,1) AND page_count > 100) AS tp
go
	
-- What about the average across all databases on your server:
sp_msforeachdb 'USE ?; SELECT DB_NAME(), AVG(tp.[LOW]) AS [Avg Low Percent]
	, AVG(tp.[High]) AS [Avg High Percent]
FROM (SELECT ((page_count * .25)/record_count *100) AS [Low]
    , ((page_count * .33)/record_count * 100) AS [High]
	FROM sys.dm_db_index_physical_stats
		(db_id(), null, null, null, ''sampled'')
	WHERE index_id IN (0,1) AND page_count > 100) AS tp'
go


------------------------------------------------------------------
-- Clean up and prep for next execution:
------------------------------------------------------------------
USE AdventureWorksDW2008
go

CREATE NONCLUSTERED INDEX IX_FactInternetSales_CustomerKey
ON [dbo].[FactInternetSales] ([CustomerKey])
go

DROP INDEX [FactInternetSales].IX_FactInternetSales_CustomerKey_INCSalesAmount
go